@@ -4,7 +4,7 @@ use 5.008;
 use strict;
 use warnings;
 
-our $VERSION = '0.03_1';
+our $VERSION = '0.07_1';
 
 =head1 NAME
 
@@ -17,9 +17,14 @@ Search::Estraier - pure perl module to use Hyper Estraier search engine
 	use Search::Estraier;
 
 	# create and configure node
-	my $node = new Search::Estraier::Node;
-	$node->set_url("http://localhost:1978/node/test");
-	$node->set_auth("admin","admin");
+	my $node = new Search::Estraier::Node(
+		url => 'http://localhost:1978/node/test',
+		user => 'admin',
+		passwd => 'admin',
+		create => 1,
+		label => 'Label for node',
+		croak_on_error => 1,
+	);
 
 	# create document
 	my $doc = new Search::Estraier::Document;
@@ -32,16 +37,19 @@ Search::Estraier - pure perl module to use Hyper Estraier search engine
 	$doc->add_text("Somewhere over the rainbow.  Way up high.");
 	$doc->add_text("There's a land that I heard of once in a lullaby.");
 
-	die "error: ", $node->status,"\n" unless ($node->put_doc($doc));
+	die "error: ", $node->status,"\n" unless (eval { $node->put_doc($doc) });
 
 =head2 Simple searcher
 
 	use Search::Estraier;
 
 	# create and configure node
-	my $node = new Search::Estraier::Node;
-	$node->set_url("http://localhost:1978/node/test");
-	$node->set_auth("admin","admin");
+	my $node = new Search::Estraier::Node(
+		url => 'http://localhost:1978/node/test',
+		user => 'admin',
+		passwd => 'admin',
+		croak_on_error => 1,
+	);
 
 	# create condition
 	my $cond = new Search::Estraier::Condition;
@@ -50,7 +58,10 @@ Search::Estraier - pure perl module to use Hyper Estraier search engine
 	$cond->set_phrase("rainbow AND lullaby");
 
 	my $nres = $node->search($cond, 0);
+
 	if (defined($nres)) {
+		print "Got ", $nres->hits, " results\n";
+
 		# for each document in results
 		for my $i ( 0 ... $nres->doc_num - 1 ) {
 			# get result document
@@ -92,7 +103,8 @@ Remove multiple whitespaces from string, as well as whitespaces at beginning or
 =cut
 
 sub _s {
-	my $text = $_[1] || return;
+	my $text = $_[1];
+	return unless defined($text);
 	$text =~ s/\s\s+/ /gs;
 	$text =~ s/^\s+//;
 	$text =~ s/\s+$//;
@@ -157,12 +169,12 @@ sub new {
 			} elsif ($line =~ m/^$/) {
 				$in_text = 1;
 				next;
-			} elsif ($line =~ m/^(.+)=(.+)$/) {
+			} elsif ($line =~ m/^(.+)=(.*)$/) {
 				$self->{attrs}->{ $1 } = $2;
 				next;
 			}
 
-			warn "draft ignored: $line\n";
+			warn "draft ignored: '$line'\n";
 		}
 	}
 
@@ -320,7 +332,8 @@ sub dump_draft {
 	my $draft;
 
 	foreach my $attr_name (sort keys %{ $self->{attrs} }) {
-		$draft .= $attr_name . '=' . $self->{attrs}->{$attr_name} . "\n";
+		next unless defined(my $v = $self->{attrs}->{$attr_name});
+		$draft .= $attr_name . '=' . $v . "\n";
 	}
 
 	if ($self->{kwords}) {
@@ -368,7 +381,7 @@ sub delete {
 
 package Search::Estraier::Condition;
 
-use Carp qw/confess croak/;
+use Carp qw/carp confess croak/;
 
 use Search::Estraier;
 our @ISA = qw/Search::Estraier/;
@@ -446,30 +459,71 @@ sub set_max {
 
 =head2 set_options
 
-  $cond->set_options( SURE => 1 );
+  $cond->set_options( 'SURE' );
+
+  $cond->set_options( qw/AGITO NOIDF SIMPLE/ );
+
+Possible options are:
+
+=over 8
+
+=item SURE
+
+check every N-gram
+
+=item USUAL
+
+check every second N-gram
+
+=item FAST
+
+check every third N-gram
+
+=item AGITO
+
+check every fourth N-gram
+
+=item NOIDF
+
+don't perform TF-IDF tuning
+
+=item SIMPLE
+
+use simplified query phrase
+
+=back
+
+Skipping N-grams will speed up search, but reduce accuracy. Every call to C<set_options> will reset previous
+options;
+
+This option changed in version C<0.04> of this module. It's backwards compatibile.
 
 =cut
 
 my $options = {
-	# check N-gram keys skipping by three
 	SURE => 1 << 0,
-	# check N-gram keys skipping by two
 	USUAL => 1 << 1,
-	# without TF-IDF tuning
 	FAST => 1 << 2,
-	# with the simplified phrase
 	AGITO => 1 << 3,
-	# check every N-gram key
 	NOIDF => 1 << 4,
-	# check N-gram keys skipping by one
 	SIMPLE => 1 << 10,
 };
 
 sub set_options {
 	my $self = shift;
-	my $option = shift;
-	confess "unknown option" unless ($options->{$option});
-	$self->{options} ||= $options->{$option};
+	my $opt = 0;
+	foreach my $option (@_) {
+		my $mask;
+		unless ($mask = $options->{$option}) {
+			if ($option eq '1') {
+				next;
+			} else {
+				croak "unknown option $option";
+			}
+		}
+		$opt += $mask;
+	}
+	$self->{options} = $opt;
 }
 
 
@@ -548,6 +602,35 @@ sub options {
 }
 
 
+=head2 set_skip
+
+Set number of skipped documents from beginning of results
+
+  $cond->set_skip(42);
+
+Similar to C<offset> in RDBMS.
+
+=cut
+
+sub set_skip {
+	my $self = shift;
+	$self->{skip} = shift;
+}
+
+=head2 skip
+
+Return skip for this condition.
+
+  print $cond->skip;
+
+=cut
+
+sub skip {
+	my $self = shift;
+	return $self->{skip};
+}
+
+
 package Search::Estraier::ResultDocument;
 
 use Carp qw/croak/;
@@ -691,6 +774,9 @@ Return number of documents
 
   print $res->doc_num;
 
+This will return real number of documents (limited by C<max>).
+If you want to get total number of hits, see C<hits>.
+
 =cut
 
 sub doc_num {
@@ -722,7 +808,7 @@ sub get_doc {
 
 Return specific hint from results.
 
-  print $rec->hint( 'VERSION' );
+  print $res->hint( 'VERSION' );
 
 Possible hints are: C<VERSION>, C<NODE>, C<HIT>, C<HINT#n>, C<DOCNUM>, C<WORDNUM>,
 C<TIME>, C<LINK#n>, C<VIEW>.
@@ -735,6 +821,37 @@ sub hint {
 	return $self->{hints}->{$key};
 }
 
+=head2 hints
+
+More perlish version of C<hint>. This one returns hash.
+
+  my %hints = $res->hints;
+
+=cut
+
+sub hints {
+	my $self = shift;
+	return $self->{hints};
+}
+
+=head2 hits
+
+Syntaxtic sugar for total number of hits for this query
+
+  print $res->hits;
+
+It's same as
+
+  print $res->hint('HIT');
+
+but shorter.
+
+=cut
+
+sub hits {
+	my $self = shift;
+	return $self->{hints}->{'HIT'} || 0;
+}
 
 package Search::Estraier::Node;
 
@@ -754,6 +871,54 @@ or optionally with C<url> as parametar
 
   my $node = new Search::HyperEstraier::Node( 'http://localhost:1978/node/test' );
 
+or in more verbose form
+
+  my $node = new Search::HyperEstraier::Node(
+  	url => 'http://localhost:1978/node/test',
+	user => 'admin',
+	passwd => 'admin'
+	create => 1,
+	label => 'optional node label',
+	debug => 1,
+	croak_on_error => 1
+  );
+
+with following arguments:
+
+=over 4
+
+=item url
+
+URL to node
+
+=item user
+
+specify username for node server authentication
+
+=item passwd
+
+password for authentication
+
+=item create
+
+create node if it doesn't exists
+
+=item label
+
+optional label for new node if C<create> is used
+
+=item debug
+
+dumps a B<lot> of debugging output
+
+=item croak_on_error
+
+very helpful during development. It will croak on all errors instead of
+silently returning C<-1> (which is convention of Hyper Estraier API in other
+languages).
+
+=back
+
 =cut
 
 sub new {
@@ -761,25 +926,43 @@ sub new {
 	my $self = {
 		pxport => -1,
 		timeout => 0,	# this used to be -1
-		dnum => -1,
-		wnum => -1,
-		size => -1.0,
 		wwidth => 480,
 		hwidth => 96,
 		awidth => 96,
 		status => -1,
 	};
+
 	bless($self, $class);
 
 	if ($#_ == 0) {
 		$self->{url} = shift;
 	} else {
-		my $args = {@_};
+		%$self = ( %$self, @_ );
+
+		$self->set_auth( $self->{user}, $self->{passwd} ) if ($self->{user});
 
-		$self->{debug} = $args->{debug};
 		warn "## Node debug on\n" if ($self->{debug});
 	}
 
+	$self->{inform} = {
+		dnum => -1,
+		wnum => -1,
+		size => -1.0,
+	};
+
+	if ($self->{create}) {
+		if (! eval { $self->name } || $@) {
+			my $name = $1 if ($self->{url} =~ m#/node/([^/]+)/*#);
+			croak "can't find node name in '$self->{url}'" unless ($name);
+			my $label = $self->{label} || $name;
+			$self->master(
+				action => 'nodeadd',
+				name => $name,
+				label => $label,
+			) || croak "can't create node $name ($label)";
+		}
+	}
+
 	$self ? return $self : return undef;
 }
 
@@ -870,7 +1053,7 @@ Add a document
 
   $node->put_doc( $document_draft ) or die "can't add document";
 
-Return true on success or false on failture.
+Return true on success or false on failure.
 
 =cut
 
@@ -878,11 +1061,15 @@ sub put_doc {
 	my $self = shift;
 	my $doc = shift || return;
 	return unless ($self->{url} && $doc->isa('Search::Estraier::Document'));
-	$self->shuttle_url( $self->{url} . '/put_doc',
+	if ($self->shuttle_url( $self->{url} . '/put_doc',
 		'text/x-estraier-draft',
 		$doc->dump_draft,
 		undef
-	) == 200;
+	) == 200) {
+		$self->_clear_info;
+		return 1;
+	}
+	return undef;
 }
 
 
@@ -901,11 +1088,15 @@ sub out_doc {
 	my $id = shift || return;
 	return unless ($self->{url});
 	croak "id must be number, not '$id'" unless ($id =~ m/^\d+$/);
-	$self->shuttle_url( $self->{url} . '/out_doc',
+	if ($self->shuttle_url( $self->{url} . '/out_doc',
 		'application/x-www-form-urlencoded',
 		"id=$id",
 		undef
-	) == 200;
+	) == 200) {
+		$self->_clear_info;
+		return 1;
+	}
+	return undef;
 }
 
 
@@ -923,11 +1114,15 @@ sub out_doc_by_uri {
 	my $self = shift;
 	my $uri = shift || return;
 	return unless ($self->{url});
-	$self->shuttle_url( $self->{url} . '/out_doc',
+	if ($self->shuttle_url( $self->{url} . '/out_doc',
 		'application/x-www-form-urlencoded',
 		"uri=" . uri_escape($uri),
 		undef
-	) == 200;
+	) == 200) {
+		$self->_clear_info;
+		return 1;
+	}
+	return undef;
 }
 
 
@@ -945,11 +1140,15 @@ sub edit_doc {
 	my $self = shift;
 	my $doc = shift || return;
 	return unless ($self->{url} && $doc->isa('Search::Estraier::Document'));
-	$self->shuttle_url( $self->{url} . '/edit_doc',
+	if ($self->shuttle_url( $self->{url} . '/edit_doc',
 		'text/x-estraier-draft',
 		$doc->dump_draft,
 		undef
-	) == 200;
+	) == 200) {
+		$self->_clear_info;
+		return 1;
+	}
+	return undef;
 }
 
 
@@ -1058,12 +1257,14 @@ Get ID of document specified by URI
 
   my $id = $node->uri_to_id( 'file:///document/uri/42' );
 
+This method won't croak, even if using C<croak_on_error>.
+
 =cut
 
 sub uri_to_id {
 	my $self = shift;
 	my $uri = shift || return;
-	return $self->_fetch_doc( uri => $uri, path => '/uri_to_id', chomp_resbody => 1 );
+	return $self->_fetch_doc( uri => $uri, path => '/uri_to_id', chomp_resbody => 1, croak_on_error => 0 );
 }
 
 
@@ -1123,6 +1324,7 @@ sub _fetch_doc {
 		'application/x-www-form-urlencoded',
 		$arg,
 		\$resbody,
+		$a->{croak_on_error},
 	);
 
 	return if ($rv != 200);
@@ -1153,8 +1355,8 @@ sub _fetch_doc {
 
 sub name {
 	my $self = shift;
-	$self->_set_info unless ($self->{name});
-	return $self->{name};
+	$self->_set_info unless ($self->{inform}->{name});
+	return $self->{inform}->{name};
 }
 
 
@@ -1166,8 +1368,8 @@ sub name {
 
 sub label {
 	my $self = shift;
-	$self->_set_info unless ($self->{label});
-	return $self->{label};
+	$self->_set_info unless ($self->{inform}->{label});
+	return $self->{inform}->{label};
 }
 
 
@@ -1179,8 +1381,8 @@ sub label {
 
 sub doc_num {
 	my $self = shift;
-	$self->_set_info if ($self->{dnum} < 0);
-	return $self->{dnum};
+	$self->_set_info if ($self->{inform}->{dnum} < 0);
+	return $self->{inform}->{dnum};
 }
 
 
@@ -1192,8 +1394,8 @@ sub doc_num {
 
 sub word_num {
 	my $self = shift;
-	$self->_set_info if ($self->{wnum} < 0);
-	return $self->{wnum};
+	$self->_set_info if ($self->{inform}->{wnum} < 0);
+	return $self->{inform}->{wnum};
 }
 
 
@@ -1205,8 +1407,8 @@ sub word_num {
 
 sub size {
 	my $self = shift;
-	$self->_set_info if ($self->{size} < 0);
-	return $self->{size};
+	$self->_set_info if ($self->{inform}->{size} < 0);
+	return $self->{inform}->{size};
 }
 
 
@@ -1239,88 +1441,32 @@ sub search {
 	);
 	return if ($rv != 200);
 
-	my (@docs, $hints);
+	my @records 	= split /--------\[.*?\]--------(?::END)?\r?\n/, $resbody;
+	my $hintsText	= splice @records, 0, 2; # starts with empty record
+	my $hints		= { $hintsText =~ m/^(.*?)\t(.*?)$/gsm };
 
-	my @lines = split(/\n/, $resbody);
-	return unless (@lines);
+	# process records
+	my $docs = [];
+	foreach my $record (@records)
+	{
+		# split into keys and snippets
+		my ($keys, $snippet) = $record =~ m/^(.*?)\n\n(.*?)$/s;
 
-	my $border = $lines[0];
-	my $isend = 0;
-	my $lnum = 1;
+		# create document hash
+		my $doc				= { $keys =~ m/^(.*?)=(.*?)$/gsm };
+		$doc->{'@keywords'}	= $doc->{keywords};
+		($doc->{keywords})	= $keys =~ m/^%VECTOR\t(.*?)$/gm;
+		$doc->{snippet}		= $snippet;
 
-	while ( $lnum <= $#lines ) {
-		my $line = $lines[$lnum];
-		$lnum++;
-
-		#warn "## $line\n";
-		if ($line && $line =~ m/^\Q$border\E(:END)*$/) {
-			$isend = $1;
-			last;
-		}
-
-		if ($line =~ /\t/) {
-			my ($k,$v) = split(/\t/, $line, 2);
-			$hints->{$k} = $v;
-		}
+		push @$docs, new Search::Estraier::ResultDocument(
+			attrs 		=> $doc,
+			uri 		=> $doc->{'@uri'},
+			snippet 	=> $snippet,
+			keywords 	=> $doc->{'keywords'},
+		);
 	}
 
-	my $snum = $lnum;
-
-	while( ! $isend && $lnum <= $#lines ) {
-		my $line = $lines[$lnum];
-		#warn "# $lnum: $line\n";
-		$lnum++;
-
-		if ($line && $line =~ m/^\Q$border\E/) {
-			if ($lnum > $snum) {
-				my $rdattrs;
-				my $rdvector;
-				my $rdsnippet;
-				
-				my $rlnum = $snum;
-				while ($rlnum < $lnum - 1 ) {
-					#my $rdline = $self->_s($lines[$rlnum]);
-					my $rdline = $lines[$rlnum];
-					$rlnum++;
-					last unless ($rdline);
-					if ($rdline =~ /^%/) {
-						$rdvector = $1 if ($rdline =~ /^%VECTOR\t(.+)$/);
-					} elsif($rdline =~ /=/) {
-						$rdattrs->{$1} = $2 if ($rdline =~ /^(.+)=(.+)$/);
-					} else {
-						confess "invalid format of response";
-					}
-				}
-				while($rlnum < $lnum - 1) {
-					my $rdline = $lines[$rlnum];
-					$rlnum++;
-					$rdsnippet .= "$rdline\n";
-				}
-				#warn Dumper($rdvector, $rdattrs, $rdsnippet);
-				if (my $rduri = $rdattrs->{'@uri'}) {
-					push @docs, new Search::Estraier::ResultDocument(
-						uri => $rduri,
-						attrs => $rdattrs,
-						snippet => $rdsnippet,
-						keywords => $rdvector,
-					);
-				}
-			}
-			$snum = $lnum;
-			#warn "### $line\n";
-			$isend = 1 if ($line =~ /:END$/);
-		}
-
-	}
-
-	if (! $isend) {
-		warn "received result doesn't have :END\n$resbody";
-		return;
-	}
-
-	#warn Dumper(\@docs, $hints);
-
-	return new Search::Estraier::NodeResult( docs => \@docs, hints => $hints );
+	return new Search::Estraier::NodeResult( docs => $docs, hints => $hints );
 }
 
 
@@ -1369,6 +1515,7 @@ sub cond_to_query {
 	push @args, 'wwidth=' . $self->{wwidth};
 	push @args, 'hwidth=' . $self->{hwidth};
 	push @args, 'awidth=' . $self->{awidth};
+	push @args, 'skip=' . $cond->{skip} if ($cond->{skip});
 
 	return join('&', @args);
 }
@@ -1391,7 +1538,9 @@ use LWP::UserAgent;
 sub shuttle_url {
 	my $self = shift;
 
-	my ($url, $content_type, $reqbody, $resbody) = @_;
+	my ($url, $content_type, $reqbody, $resbody, $croak_on_error) = @_;
+
+	$croak_on_error = $self->{croak_on_error} unless defined($croak_on_error);
 
 	$self->{status} = -1;
 
@@ -1418,7 +1567,7 @@ sub shuttle_url {
 
 	$req->headers->header( 'Host' => $url->host . ":" . $url->port );
 	$req->headers->header( 'Connection', 'close' );
-	$req->headers->header( 'Authorization', 'Basic ' . $self->{auth} );
+	$req->headers->header( 'Authorization', 'Basic ' . $self->{auth} ) if ($self->{auth});
 	$req->content_type( $content_type );
 
 	warn $req->headers->as_string,"\n" if ($self->{debug});
@@ -1432,10 +1581,16 @@ sub shuttle_url {
 
 	warn "## response status: ",$res->status_line,"\n" if ($self->{debug});
 
-	return -1 if (! $res->is_success);
-
 	($self->{status}, $self->{status_message}) = split(/\s+/, $res->status_line, 2);
 
+	if (! $res->is_success) {
+		if ($croak_on_error) {
+			croak("can't get $url: ",$res->status_line);
+		} else {
+			return -1;
+		}
+	}
+
 	$$resbody .= $res->content;
 
 	warn "## response body:\n$$resbody\n" if ($resbody && $self->{debug});
@@ -1507,7 +1662,7 @@ sub set_user {
 	croak "mode must be number, not '$mode'" unless ($mode =~ m/^\d+$/);
 
 	$self->shuttle_url( $self->{url} . '/_set_user',
-		'text/plain',
+		'application/x-www-form-urlencoded',
 		'name=' . uri_escape($name) . '&mode=' . $mode,
 		undef
 	) == 200;
@@ -1534,13 +1689,193 @@ sub set_link {
 	my $reqbody = 'url=' . uri_escape($url) . '&label=' . uri_escape($label);
 	$reqbody .= '&credit=' . $credit if ($credit > 0);
 
-	$self->shuttle_url( $self->{url} . '/_set_link',
-		'text/plain',
+	if ($self->shuttle_url( $self->{url} . '/_set_link',
+		'application/x-www-form-urlencoded',
 		$reqbody,
 		undef
-	) == 200;
+	) == 200) {
+		# refresh node info after adding link
+		$self->_clear_info;
+		return 1;
+	}
+	return undef;
 }
 
+=head2 admins
+
+ my @admins = @{ $node->admins };
+
+Return array of users with admin rights on node
+
+=cut
+
+sub admins {
+	my $self = shift;
+	$self->_set_info unless ($self->{inform}->{name});
+	return $self->{inform}->{admins};
+}
+
+=head2 guests
+
+ my @guests = @{ $node->guests };
+
+Return array of users with guest rights on node
+
+=cut
+
+sub guests {
+	my $self = shift;
+	$self->_set_info unless ($self->{inform}->{name});
+	return $self->{inform}->{guests};
+}
+
+=head2 links
+
+ my $links = @{ $node->links };
+
+Return array of links for this node
+
+=cut
+
+sub links {
+	my $self = shift;
+	$self->_set_info unless ($self->{inform}->{name});
+	return $self->{inform}->{links};
+}
+
+=head2 master
+
+Set actions on Hyper Estraier node master (C<estmaster> process)
+
+  $node->master(
+  	action => 'sync'
+  );
+
+All available actions are documented in
+L<http://hyperestraier.sourceforge.net/nguide-en.html#protocol>
+
+=cut
+
+my $estmaster_rest = {
+	shutdown => {
+		status => 202,
+	},
+	sync => {
+		status => 202,
+	},
+	backup => {
+		status => 202,
+	},
+	userlist => {
+		status => 200,
+		returns => [ qw/name passwd flags fname misc/ ],
+	},
+	useradd => {
+		required => [ qw/name passwd flags/ ],
+		optional => [ qw/fname misc/ ],
+		status => 200,
+	},
+	userdel => {
+		required => [ qw/name/ ],
+		status => 200,
+	},
+	nodelist => {
+		status => 200,
+		returns => [ qw/name label doc_num word_num size/ ],
+	},
+	nodeadd => {
+		required => [ qw/name/ ],
+		optional => [ qw/label/ ],
+		status => 200,
+	},
+	nodedel => {
+		required => [ qw/name/ ],
+		status => 200,
+	},
+	nodeclr => {
+		required => [ qw/name/ ],
+		status => 200,
+	},
+	nodertt => {
+		status => 200,	
+	},
+};
+
+sub master {
+	my $self = shift;
+
+	my $args = {@_};
+
+	# have action?
+	my $action = $args->{action} || croak "need action, available: ",
+		join(", ",keys %{ $estmaster_rest });
+
+	# check if action is valid
+	my $rest = $estmaster_rest->{$action};
+	croak "action '$action' is not supported, available actions: ",
+		join(", ",keys %{ $estmaster_rest }) unless ($rest);
+
+	croak "BUG: action '$action' needs return status" unless ($rest->{status});
+
+	my @args;
+
+	if ($rest->{required} || $rest->{optional}) {
+
+		map {
+			croak "need parametar '$_' for action '$action'" unless ($args->{$_});
+			push @args, $_ . '=' . uri_escape( $args->{$_} );
+		} ( @{ $rest->{required} } );
+
+		map {
+			push @args, $_ . '=' . uri_escape( $args->{$_} ) if ($args->{$_});
+		} ( @{ $rest->{optional} } );
+
+	}
+
+	my $uri = new URI( $self->{url} );
+
+	my $resbody;
+
+	my $status = $self->shuttle_url(
+		'http://' . $uri->host_port . '/master?action=' . $action ,
+		'application/x-www-form-urlencoded',
+		join('&', @args),
+		\$resbody,
+		1,
+	) or confess "shuttle_url failed";
+
+	if ($status == $rest->{status}) {
+
+		# refresh node info after sync
+		$self->_clear_info if ($action eq 'sync' || $action =~ m/^node(?:add|del|clr)$/);
+
+		if ($rest->{returns} && wantarray) {
+
+			my @results;
+			my $fields = $#{$rest->{returns}};
+
+			foreach my $line ( split(/[\r\n]/,$resbody) ) {
+				my @e = split(/\t/, $line, $fields + 1);
+				my $row;
+				foreach my $i ( 0 .. $fields) {
+					$row->{ $rest->{returns}->[$i] } = $e[ $i ];
+				}
+				push @results, $row;
+			}
+
+			return @results;
+
+		} elsif ($resbody) {
+			chomp $resbody;
+			return $resbody;
+		} else {
+			return 0E0;
+		}
+	}
+
+	carp "expected status $rest->{status}, but got $status";
+	return undef;
+}
 
 =head1 PRIVATE METHODS
 
@@ -1569,12 +1904,50 @@ sub _set_info {
 
 	return if ($rv != 200 || !$resbody);
 
-	# it seems that response can have multiple line endings
-	$resbody =~ s/[\r\n]+$//;
+	my @lines = split(/[\r\n]/,$resbody);
 
-	( $self->{name}, $self->{label}, $self->{dnum}, $self->{wnum}, $self->{size} ) =
-		split(/\t/, $resbody, 5);
+	$self->_clear_info;
 
+	( $self->{inform}->{name}, $self->{inform}->{label}, $self->{inform}->{dnum},
+		$self->{inform}->{wnum}, $self->{inform}->{size} ) = split(/\t/, shift @lines, 5);
+
+	return $resbody unless (@lines);
+
+	shift @lines;
+
+	while(my $admin = shift @lines) {
+		push @{$self->{inform}->{admins}}, $admin;
+	}
+
+	while(my $guest = shift @lines) {
+		push @{$self->{inform}->{guests}}, $guest;
+	}
+
+	while(my $link = shift @lines) {
+		push @{$self->{inform}->{links}}, $link;
+	}
+
+	return $resbody;
+
+}
+
+=head2 _clear_info
+
+Clear information for node
+
+  $node->_clear_info;
+
+On next call to C<name>, C<label>, C<doc_num>, C<word_num> or C<size> node
+info will be fetch again from Hyper Estraier.
+
+=cut
+sub _clear_info {
+	my $self = shift;
+	$self->{inform} = {
+		dnum => -1,
+		wnum => -1,
+		size => -1.0,
+	};
 }
 
 ###
@@ -1593,6 +1966,7 @@ Hyper Estraier Ruby interface on which this module is based.
 
 Dobrica Pavlinusic, E<lt>dpavlin@rot13.orgE<gt>
 
+Robert Klep E<lt>robert@klep.nameE<gt> contributed refactored search code
 
 =head1 COPYRIGHT AND LICENSE
 
